# -------------------------------
# SPDX-License-Identifier: MIT
#
# Copyright © 2020 Andrew Hardin
# Copyright © 2021 Daniel Wolf <<nephatrine@gmail.com>>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the “Software”), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# -------------------------------

# -------------------------------
# Adapted from Andrew Hardin's cmake-git-version-tracking script. https://github.com/andrew-hardin/cmake-git-version-tracking
# -------------------------------

if(NOT DEFINED GIT_EXECUTABLE)
	find_program(GIT_EXECUTABLE NAMES git)
endif()

set(NX_GITWATCH_FILE "${CMAKE_CURRENT_LIST_FILE}")
set(NX_GITWATCH_VARS "GIT_RETRIEVED_STATE" "GIT_BRANCH_NAME" "GIT_COMMIT_HASH" "GIT_IS_DIRTY" "GIT_COMMIT_DATE" "GIT_COMMIT_TIME")

macro(_nx_git_command)
	set(nGitResult 255)
	set(sGitOutput "")

	if(GIT_EXECUTABLE)
		execute_process(
			COMMAND "${GIT_EXECUTABLE}" ${ARGV}
			WORKING_DIRECTORY "${sWorkingDirectory}"
			RESULT_VARIABLE nGitResult
			OUTPUT_VARIABLE sGitOutput
			ERROR_QUIET OUTPUT_STRIP_TRAILING_WHITESPACE)
	endif()

	if(NOT nGitResult EQUAL 0)
		set(ENV{GIT_RETRIEVED_STATE} "false")
	endif()
endmacro()

function(nx_git_update sWorkingDirectory)
	set(ENV{GIT_RETRIEVED_STATE} "true")

	_nx_git_command(rev-parse --abbrev-ref HEAD)
	if(NOT nGitResult EQUAL 0 OR sGitOutput STREQUAL "")
		set(ENV{GIT_BRANCH_NAME} "unknown")
	else()
		set(ENV{GIT_BRANCH_NAME} "${sGitOutput}")
	endif()

	_nx_git_command(log -1 --format=%h)
	if(NOT nGitResult EQUAL 0 OR sGitOutput STREQUAL "")
		set(ENV{GIT_COMMIT_HASH} "unknown")
	else()
		set(ENV{GIT_COMMIT_HASH} "${sGitOutput}")
	endif()

	_nx_git_command(log -1 --format=%cd --date=format:%Y%m%d)
	if(NOT nGitResult EQUAL 0 OR sGitOutput STREQUAL "")
		set(ENV{GIT_COMMIT_DATE} "unknown")
	else()
		set(ENV{GIT_COMMIT_DATE} "${sGitOutput}")
	endif()

	_nx_git_command(log -1 --format=%cd --date=format:%H%M%S)
	if(NOT nGitResult EQUAL 0 OR sGitOutput STREQUAL "")
		set(ENV{GIT_COMMIT_TIME} "unknown")
	else()
		set(ENV{GIT_COMMIT_TIME} "${sGitOutput}")
	endif()

	_nx_git_command(status --porcelain -uno)
	if(NOT nGitResult EQUAL 0 OR sGitOutput STREQUAL "")
		set(ENV{GIT_IS_DIRTY} "false")
	else()
		set(ENV{GIT_IS_DIRTY} "true")
	endif()
endfunction()

function(nx_git_watcher)
	if(NOT DEFINED _NX_GIT_INTERNAL)
		return()
	endif()

	nx_git_update("${NX_GIT_SOURCE_DIR}")

	if(EXISTS "${NX_GIT_FILE_INPUT}")
		file(READ "${NX_GIT_FILE_INPUT}" sGitInput)
		string(SHA256 sGitState "${sGitInput}")
	endif()

	foreach(vGitWatch ${NX_GITWATCH_VARS})
		string(SHA256 sGitState "${sGitState}$ENV{${vGitWatch}}")
	endforeach()

	set(bGitChanged ON)
	if(EXISTS "${NX_GIT_FILE_STATE}")
		file(READ "${NX_GIT_FILE_STATE}" sGitPrevious)
		if(sGitPrevious STREQUAL "${sGitState}")
			set(bGitChanged OFF)
		endif()
	endif()

	if(bGitChanged)
		file(WRITE "${NX_GIT_FILE_STATE}" "${sGitState}")
	endif()

	if(bGitChanged OR NOT EXISTS "${NX_GIT_FILE_OUTPUT}")
		foreach(vGitWatch ${NX_GITWATCH_VARS})
			set(NX_${vGitWatch} $ENV{${vGitWatch}})
		endforeach()
		configure_file("${NX_GIT_FILE_INPUT}" "${NX_GIT_FILE_OUTPUT}")
	endif()
endfunction()

function(nx_git_check)
	nx_git_update("${CMAKE_CURRENT_SOURCE_DIR}")
	foreach(vGitWatch ${NX_GITWATCH_VARS})
		set(NX_${vGitWatch}
			$ENV{${vGitWatch}}
			PARENT_SCOPE)
	endforeach()
endfunction()

function(nx_git_configure sGitInput sGitOutput)
	get_filename_component(sGitTarget "${sGitInput}" NAME)
	string(MAKE_C_IDENTIFIER "${sGitTarget}" sGitTarget)
	add_custom_target(
		${sGitTarget} ALL
		BYPRODUCTS "${sGitOutput}"
		DEPENDS "${sGitInput}"
		COMMENT "[GitWatcher] Retrieving Git State"
		COMMAND
			${CMAKE_COMMAND} -D_NX_GIT_INTERNAL=ON -DGIT_EXECUTABLE=${GIT_EXECUTABLE} -DNX_GIT_SOURCE_DIR="${CMAKE_CURRENT_SOURCE_DIR}"
			-DNX_GIT_FILE_INPUT=${sGitInput} -DNX_GIT_FILE_OUTPUT="${sGitOutput}"
			-DNX_GIT_FILE_STATE="${CMAKE_CURRENT_BINARY_DIR}/git.state" -P "${NX_GITWATCH_FILE}")
	configure_file("${sGitInput}" "${sGitOutput}")
endfunction()

nx_git_watcher()
