# -------------------------------
# SPDX-License-Identifier: ISC
#
# Copyright © 2021 Daniel Wolf <<nephatrine@gmail.com>>
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
# AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
# INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
# LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
# OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
# PERFORMANCE OF THIS SOFTWARE.
# -------------------------------

include(NXIncludeGuard)
include(NXScopePropagate)

nx_include_guard()

# ===================================================================

set(NXGenerateVersionHeader_DIRECTORY "${CMAKE_CURRENT_LIST_DIR}")

function(nx_generate_version_header sBaseName)
	nx_scope_initialize()

	if(DEFINED ${sBaseName}_VERSION)
		set(GVH_VERSION ${${sBaseName}_VERSION})
	elseif(DEFINED PROJECT_VERSION)
		set(GVH_VERSION ${PROJECT_VERSION})
	else()
		set(GVH_VERSION "0.0.0")
	endif()

	string(REPLACE "." ";" lsVersionComponents "${GVH_VERSION}")
	list(LENGTH lsVersionComponents nVersionComponents)

	if(nVersionComponents GREATER 0 AND NOT DEFINED sProjectAPIMAJOR)
		list(GET lsVersionComponents 0 GVH_VERSION_MAJOR)
	else()
		set(GVH_VERSION_MAJOR 0)
	endif()
	if(nVersionComponents GREATER 1 AND NOT DEFINED sProjectAPIMINOR)
		list(GET lsVersionComponents 1 GVH_VERSION_MINOR)
	else()
		set(GVH_VERSION_MINOR 0)
	endif()
	if(nVersionComponents GREATER 2 AND NOT DEFINED sProjectAPIPATCH)
		list(GET lsVersionComponents 2 GVH_VERSION_PATCH)
	else()
		set(GVH_VERSION_PATCH 0)
	endif()

	set(GVH_VERSION_HEX "${GVH_VERSION_MAJOR}*65536 + ${GVH_VERSION_MINOR}*256 + ${GVH_VERSION_PATCH}")
	if(CMAKE_VERSION VERSION_LESS 3.13)
		math(EXPR GVH_VERSION_HEX "${GVH_VERSION_HEX}")
	else()
		math(EXPR GVH_VERSION_HEX "${GVH_VERSION_HEX}" OUTPUT_FORMAT HEXADECIMAL)
	endif()

	string(TOUPPER "${sBaseName}" sBaseName)
	string(MAKE_C_IDENTIFIER "${sBaseName}" sBaseName)

	set(GVH_BASE_NAME ${sBaseName})
	set(GVH_QUERY_GIT 0)

	unset(GVH_CUSTOM_CONTENT)
	unset(GVH_CUSTOM_CONTENT_FROM_VARIABLE)
	unset(GVH_INCLUDE_GUARD_NAME)
	unset(GVH_PREFIX_NAME)
	unset(GVH_VERSION_FILE_NAME)
	unset(GVH_VERSION_MACRO_NAME)

	set(lsParseModes BASE_NAME VERSION_MACRO_NAME VERSION_FILE_NAME INCLUDE_GUARD_NAME PREFIX_NAME CUSTOM_CONTENT_FROM_VARIABLE)
	set(lsParseToggles QUERY_GIT)
	set(sParseMode "VERSION_FILE_NAME")

	foreach(sArgument ${ARGN})
		if("${sArgument}" IN_LIST lsParseModes)
			set(sParseMode "${sArgument}")
		elseif("${sArgument}" IN_LIST lsParseToggles)
			set(GVH_${sArgument} 1)
		else()
			set(GVH_${sParseMode} "${sArgument}")
		endif()
	endforeach()

	if(NOT DEFINED GVH_VERSION_FILE_NAME)
		string(TOLOWER "${GVH_BASE_NAME}_version.h" GVH_VERSION_FILE_NAME)
	endif()
	if(NOT DEFINED GVH_INCLUDE_GUARD_NAME)
		get_filename_component(GVH_INCLUDE_GUARD_NAME "${GVH_VERSION_FILE_NAME}" NAME)
		string(TOUPPER "${CMAKE_PROJECT_NAME}" GVH_INCLUDE_GUARD_PREFIX)
		string(TOUPPER "${GVH_INCLUDE_GUARD_NAME}" GVH_INCLUDE_GUARD_NAME)
		if(NOT GVH_INCLUDE_GUARD_NAME MATCHES "^${GVH_INCLUDE_GUARD_PREFIX}")
			set(GVH_INCLUDE_GUARD_NAME "${GVH_INCLUDE_GUARD_PREFIX}_${GVH_INCLUDE_GUARD_NAME}")
		endif()
		string(MAKE_C_IDENTIFIER "${GVH_INCLUDE_GUARD_NAME}" GVH_INCLUDE_GUARD_NAME)
		set(GVH_INCLUDE_GUARD_NAME "${GVH_PREFIX_NAME}${GVH_INCLUDE_GUARD_NAME}")
	endif()

	if(NOT DEFINED GVH_VERSION_MACRO_NAME)
		set(GVH_VERSION_MACRO_NAME "${GVH_PREFIX_NAME}${GVH_BASE_NAME}_VERSION")
	endif()

	if(DEFINED GVH_CUSTOM_CONTENT_FROM_VARIABLE)
		if(DEFINED ${GVH_CUSTOM_CONTENT_FROM_VARIABLE})
			set(GVH_CUSTOM_CONTENT "${${GVH_CUSTOM_CONTENT_FROM_VARIABLE}}")
		endif()
	endif()

	if(NOT DEFINED _CURRENT_YEAR)
		string(TIMESTAMP _CURRENT_YEAR "%Y")
		string(TIMESTAMP _CURRENT_DATE "%Y%m%d")
	endif()

	if(GVH_QUERY_GIT)
		if(NOT DEFINED NX_GIT_WATCH_VARS)
			include(GitWatch)
		endif()
		configure_file("${NXGenerateVersionHeader_DIRECTORY}/NXGenerateVersionHeaderGit.configure" "${GVH_VERSION_FILE_NAME}.configure"
						@ONLY)
		nx_git_configure("${CMAKE_CURRENT_BINARY_DIR}/${GVH_VERSION_FILE_NAME}.configure" "${GVH_VERSION_FILE_NAME}")
	else()
		configure_file("${NXGenerateVersionHeader_DIRECTORY}/NXGenerateVersionHeader.configure" "${GVH_VERSION_FILE_NAME}")
	endif()

	nx_scope_propagate()
endfunction()
