# -------------------------------
# SPDX-License-Identifier: ISC
#
# Copyright © 2021 Daniel Wolf <<nephatrine@gmail.com>>
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
# AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
# INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
# LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
# OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
# PERFORMANCE OF THIS SOFTWARE.
# -------------------------------

include(NXIncludeGuard)
include(NXScopePropagate)

nx_include_guard()

# ===================================================================

function(nx_identify_build)
	nx_scope_initialize()

	get_property(NX_MULTI_CONFIG GLOBAL PROPERTY GENERATOR_IS_MULTI_CONFIG)
	if(NOT NX_MULTI_CONFIG AND (NOT DEFINED CMAKE_BUILD_TYPE OR NOT CMAKE_BUILD_TYPE))
		list(APPEND _NX_PROPAGATE_GLOBAL "CMAKE_BUILD_TYPE")
		set(CMAKE_BUILD_TYPE
			"RelWithDebInfo"
			CACHE STRING "" FORCE)
	endif()

	string(TOUPPER "${CMAKE_BUILD_TYPE}" NX_TARGET_BUILD)

	set(NX_TARGET_BUILD_DEBUG OFF)
	set(NX_TARGET_BUILD_MINSIZE OFF)
	set(NX_TARGET_BUILD_RELEASE OFF)
	set(NX_TARGET_BUILD_MAXSPEED OFF)
	set(NX_TARGET_BUILD_MULTI OFF)

	list(
		APPEND
		_NX_PROPAGATE_SET
		"NX_TARGET_BUILD_DEBUG"
		"NX_TARGET_BUILD_MINSIZE"
		"NX_TARGET_BUILD_RELEASE"
		"NX_TARGET_BUILD_MAXSPEED"
		"NX_TARGET_BUILD_MULTI"
		"NX_TARGET_BUILD_STRING"
		"NX_TARGET_BUILD_MACRO")
	list(APPEND _NX_PROPAGATE_SORT "NX_TARGET_BUILD_MACRO")

	if(NOT NX_MULTI_CONFIG)
		if(NX_TARGET_BUILD STREQUAL "DEBUG")
			set(NX_TARGET_BUILD_STRING "Debug")
			set(NX_TARGET_BUILD_DEBUG ON)
		elseif(NX_TARGET_BUILD STREQUAL "MINSIZEREL")
			set(NX_TARGET_BUILD_STRING "MinSize")
			set(NX_TARGET_BUILD_MINSIZE ON)
		elseif(NX_TARGET_BUILD STREQUAL "RELWITHDEBINFO")
			set(NX_TARGET_BUILD_STRING "Release")
			set(NX_TARGET_BUILD_RELEASE ON)
		elseif(NX_TARGET_BUILD STREQUAL "RELEASE")
			set(NX_TARGET_BUILD_STRING "MaxSpeed")
			set(NX_TARGET_BUILD_MAXSPEED ON)
		else()
			message(FATAL_ERROR "Unsupported Build Type ('${NX_TARGET_BUILD}')")
		endif()
		string(TOUPPER "KYUII_BUILD_${NX_TARGET_BUILD_STRING}" NX_TARGET_BUILD_MACRO)
		list(APPEND NX_TARGET_BUILD_MACRO KYUII_BUILD_STRING="${NX_TARGET_BUILD_STRING}")
	else()
		set(NX_TARGET_BUILD_STRING "Configurable")
		set(NX_TARGET_BUILD_MULTI ON)
		set(NX_TARGET_BUILD_MACRO
			$<$<CONFIG:Debug>:KYUII_BUILD_DEBUG>
			$<$<CONFIG:MinSizeRel>:KYUII_BUILD_MINSIZE>
			$<$<CONFIG:RelWithDebInfo>:KYUII_BUILD_RELEASE>
			$<$<CONFIG:Release>:KYUII_BUILD_MAXSPEED>
			$<$<CONFIG:Debug>:KYUII_BUILD_STRING="Debug">
			$<$<CONFIG:MinSizeRel>:KYUII_BUILD_STRING="MinSize">
			$<$<CONFIG:RelWithDebInfo>:KYUII_BUILD_STRING="Release">
			$<$<CONFIG:Release>:KYUII_BUILD_STRING="MaxSpeed">)
	endif()

	nx_scope_propagate()
endfunction()

function(nx_identify_architecture)
	nx_scope_initialize()

	if(DEFINED CMAKE_VS_PLATFORM_NAME)
		string(TOUPPER "${CMAKE_VS_PLATFORM_NAME}" NX_TARGET_ARCHITECTURE)
	else()
		string(TOUPPER "${CMAKE_SYSTEM_PROCESSOR}" NX_TARGET_ARCHITECTURE)
	endif()
	string(TOUPPER "${CMAKE_HOST_SYSTEM_PROCESSOR}" NX_HOST_ARCHITECTURE)

	if(DEFINED CMAKE_SIZEOF_VOID_P AND CMAKE_SIZEOF_VOID_P GREATER 0)
		math(EXPR NX_TARGET_BITS "${CMAKE_SIZEOF_VOID_P} * 8")
	else()
		set(NX_TARGET_BITS 0)
	endif()

	set(NX_ARCHITECTURES_AMD64 "AMD64" "X64" "X86_64" "X86_64H")
	set(NX_ARCHITECTURES_ARM64 "AARCH64" "ARM64")
	set(NX_ARCHITECTURES_ARMV7 "ARM")
	set(NX_ARCHITECTURES_IA32 "I386" "I486" "I586" "I686" "IA32" "WIN32" "X86")

	set(NX_HOST_ARCHITECTURE_AMD64 OFF)
	set(NX_HOST_ARCHITECTURE_ARM64 OFF)
	set(NX_HOST_ARCHITECTURE_ARMV7 OFF)
	set(NX_HOST_ARCHITECTURE_IA32 OFF)

	set(NX_TARGET_ARCHITECTURE_AMD64 OFF)
	set(NX_TARGET_ARCHITECTURE_ARM64 OFF)
	set(NX_TARGET_ARCHITECTURE_ARMV7 OFF)
	set(NX_TARGET_ARCHITECTURE_IA32 OFF)

	set(NX_TARGET_ARCHITECTURE_CROSS OFF)
	set(NX_TARGET_ARCHITECTURE_NATIVE OFF)

	list(
		APPEND
		_NX_PROPAGATE_SET
		"NX_HOST_ARCHITECTURE_AMD64"
		"NX_HOST_ARCHITECTURE_ARM64"
		"NX_HOST_ARCHITECTURE_ARMV7"
		"NX_HOST_ARCHITECTURE_IA32"
		"NX_TARGET_ARCHITECTURE_AMD64"
		"NX_TARGET_ARCHITECTURE_ARM64"
		"NX_TARGET_ARCHITECTURE_ARMV7"
		"NX_TARGET_ARCHITECTURE_IA32"
		"NX_TARGET_ARCHITECTURE_CROSS"
		"NX_TARGET_ARCHITECTURE_NATIVE"
		"NX_TARGET_ARCHITECTURE_STRING"
		"NX_TARGET_ARCHITECTURE_MACRO")
	list(APPEND _NX_PROPAGATE_SORT "NX_TARGET_ARCHITECTURE_MACRO")

	if("${NX_HOST_ARCHITECTURE}" IN_LIST NX_ARCHITECTURES_AMD64)
		set(NX_HOST_ARCHITECTURE_AMD64 ON)
	elseif("${NX_HOST_ARCHITECTURE}" IN_LIST NX_ARCHITECTURES_ARM64)
		set(NX_HOST_ARCHITECTURE_ARM64 ON)
	elseif("${NX_HOST_ARCHITECTURE}" IN_LIST NX_ARCHITECTURES_ARMV7)
		set(NX_HOST_ARCHITECTURE_ARMV7 ON)
	elseif("${NX_HOST_ARCHITECTURE}" IN_LIST NX_ARCHITECTURES_IA32)
		set(NX_HOST_ARCHITECTURE_IA32 ON)
	else()
		message(WARNING "Unknown Host Architecture ('${NX_HOST_ARCHITECTURE}')")
	endif()

	list(APPEND NX_ARCHITECTURES_IA32 ${NX_ARCHITECTURES_AMD64})
	list(APPEND NX_ARCHITECTURES_ARMV7 ${NX_ARCHITECTURES_ARM64})

	if("${NX_TARGET_ARCHITECTURE}" IN_LIST NX_ARCHITECTURES_AMD64 AND NX_TARGET_BITS EQUAL 64)
		set(NX_TARGET_ARCHITECTURE_STRING "AMD64")
		set(NX_TARGET_ARCHITECTURE_AMD64 ON)
		if(NX_HOST_ARCHITECTURE_AMD64)
			set(NX_TARGET_ARCHITECTURE_NATIVE ON)
		endif()
	elseif("${NX_TARGET_ARCHITECTURE}" IN_LIST NX_ARCHITECTURES_ARM64 AND NX_TARGET_BITS EQUAL 64)
		set(NX_TARGET_ARCHITECTURE_STRING "ARM64")
		set(NX_TARGET_ARCHITECTURE_ARM64 ON)
		if(NX_HOST_ARCHITECTURE_ARM64)
			set(NX_TARGET_ARCHITECTURE_NATIVE ON)
		endif()
	elseif("${NX_TARGET_ARCHITECTURE}" IN_LIST NX_ARCHITECTURES_ARMV7 AND NX_TARGET_BITS EQUAL 32)
		set(NX_TARGET_ARCHITECTURE_STRING "ARMv7")
		set(NX_TARGET_ARCHITECTURE_ARMV7 ON)
		if(NX_HOST_ARCHITECTURE_ARMV7 OR NX_HOST_ARCHITECTURE_ARM64)
			set(NX_TARGET_ARCHITECTURE_NATIVE ON)
		endif()
	elseif("${NX_TARGET_ARCHITECTURE}" IN_LIST NX_ARCHITECTURES_IA32 AND NX_TARGET_BITS EQUAL 32)
		set(NX_TARGET_ARCHITECTURE_STRING "IA32")
		set(NX_TARGET_ARCHITECTURE_IA32 ON)
		if(NX_HOST_ARCHITECTURE_IA32 OR NX_HOST_ARCHITECTURE_AMD64)
			set(NX_TARGET_ARCHITECTURE_NATIVE ON)
		endif()
	else()
		message(FATAL_ERROR "Unsupported Architecture Type ('${NX_TARGET_ARCHITECTURE;${NX_TARGET_BITS}-Bit}')")
	endif()

	string(TOUPPER "KYUII_ARCHITECTURE_${NX_TARGET_ARCHITECTURE_STRING}" NX_TARGET_ARCHITECTURE_MACRO)
	list(APPEND NX_TARGET_ARCHITECTURE_MACRO KYUII_ARCHITECTURE_${NX_TARGET_BITS}BIT KYUII_ARCHITECTURE_BITS=${NX_TARGET_BITS}
			KYUII_ARCHITECTURE_STRING="${NX_TARGET_ARCHITECTURE_STRING}")

	if(NOT NX_TARGET_ARCHITECTURE_NATIVE)
		set(NX_TARGET_ARCHITECTURE_CROSS ON)
	endif()

	nx_scope_propagate()
endfunction()

function(nx_identify_platform)
	nx_scope_initialize()

	string(TOUPPER "${CMAKE_SYSTEM_NAME}" NX_TARGET_PLATFORM)
	string(TOUPPER "${CMAKE_HOST_SYSTEM_NAME}" NX_HOST_PLATFORM)

	set(NX_PLATFORMS_ANDROID "ANDROID")
	set(NX_PLATFORMS_DARWIN "DARWIN")
	set(NX_PLATFORMS_FREEBSD "FREEBSD")
	set(NX_PLATFORMS_HAIKU "HAIKU")
	set(NX_PLATFORMS_LINUX "LINUX")
	set(NX_PLATFORMS_MSDOS "MSDOS")
	set(NX_PLATFORMS_SOLARIS "SUNOS")
	set(NX_PLATFORMS_WINDOWS "WINDOWS")

	set(NX_HOST_PLATFORM_DARWIN OFF)
	set(NX_HOST_PLATFORM_FREEBSD OFF)
	set(NX_HOST_PLATFORM_HAIKU OFF)
	set(NX_HOST_PLATFORM_LINUX OFF)
	set(NX_HOST_PLATFORM_SOLARIS OFF)
	set(NX_HOST_PLATFORM_WINDOWS OFF)

	set(NX_TARGET_PLATFORM_ANDROID OFF)
	set(NX_TARGET_PLATFORM_DARWIN OFF)
	set(NX_TARGET_PLATFORM_FREEBSD OFF)
	set(NX_TARGET_PLATFORM_HAIKU OFF)
	set(NX_TARGET_PLATFORM_LINUX OFF)
	set(NX_TARGET_PLATFORM_MSDOS OFF)
	set(NX_TARGET_PLATFORM_SOLARIS OFF)
	set(NX_TARGET_PLATFORM_WINDOWS OFF)

	set(NX_TARGET_PLATFORM_CROSS OFF)
	set(NX_TARGET_PLATFORM_NATIVE OFF)
	set(NX_TARGET_PLATFORM_POSIX OFF)

	list(
		APPEND
		_NX_PROPAGATE_SET
		"NX_HOST_PLATFORM_DARWIN"
		"NX_HOST_PLATFORM_FREEBSD"
		"NX_HOST_PLATFORM_HAIKU"
		"NX_HOST_PLATFORM_LINUX"
		"NX_HOST_PLATFORM_SOLARIS"
		"NX_HOST_PLATFORM_WINDOWS"
		"NX_TARGET_PLATFORM_ANDROID"
		"NX_TARGET_PLATFORM_DARWIN"
		"NX_TARGET_PLATFORM_FREEBSD"
		"NX_TARGET_PLATFORM_HAIKU"
		"NX_TARGET_PLATFORM_LINUX"
		"NX_TARGET_PLATFORM_MSDOS"
		"NX_TARGET_PLATFORM_SOLARIS"
		"NX_TARGET_PLATFORM_WINDOWS"
		"NX_TARGET_PLATFORM_CROSS"
		"NX_TARGET_PLATFORM_NATIVE"
		"NX_TARGET_PLATFORM_POSIX"
		"NX_TARGET_PLATFORM_STRING"
		"NX_TARGET_PLATFORM_MACRO")
	list(APPEND _NX_PROPAGATE_SORT "NX_TARGET_PLATFORM_MACRO")

	if("${NX_HOST_PLATFORM}" IN_LIST NX_PLATFORMS_DARWIN)
		set(NX_HOST_PLATFORM_DARWIN ON)
	elseif("${NX_HOST_PLATFORM}" IN_LIST NX_PLATFORMS_FREEBSD)
		set(NX_HOST_PLATFORM_FREEBSD ON)
	elseif("${NX_HOST_PLATFORM}" IN_LIST NX_PLATFORMS_HAIKU)
		set(NX_HOST_PLATFORM_HAIKU ON)
	elseif("${NX_HOST_PLATFORM}" IN_LIST NX_PLATFORMS_LINUX)
		set(NX_HOST_PLATFORM_LINUX ON)
	elseif("${NX_HOST_PLATFORM}" IN_LIST NX_PLATFORMS_SOLARIS)
		set(NX_HOST_PLATFORM_SOLARIS ON)
	elseif("${NX_HOST_PLATFORM}" IN_LIST NX_PLATFORMS_WINDOWS)
		set(NX_HOST_PLATFORM_WINDOWS ON)
	else()
		message(WARNING "Unknown Host Platform ('${NX_HOST_PLATFORM}')")
	endif()

	if("${NX_TARGET_PLATFORM}" IN_LIST NX_PLATFORMS_ANDROID)
		set(NX_TARGET_PLATFORM_STRING "Android")
		set(NX_TARGET_PLATFORM_ANDROID ON)
		set(NX_TARGET_PLATFORM_LINUX ON)
		set(NX_TARGET_PLATFORM_POSIX ON)
	elseif("${NX_TARGET_PLATFORM}" IN_LIST NX_PLATFORMS_DARWIN)
		set(NX_TARGET_PLATFORM_STRING "Darwin")
		set(NX_TARGET_PLATFORM_DARWIN ON)
		set(NX_TARGET_PLATFORM_POSIX ON)
		if(NX_HOST_PLATFORM_DARWIN)
			set(NX_TARGET_PLATFORM_NATIVE ON)
		endif()
	elseif("${NX_TARGET_PLATFORM}" IN_LIST NX_PLATFORMS_FREEBSD)
		set(NX_TARGET_PLATFORM_STRING "FreeBSD")
		set(NX_TARGET_PLATFORM_FREEBSD ON)
		set(NX_TARGET_PLATFORM_POSIX ON)
		if(NX_HOST_PLATFORM_FREEBSD)
			set(NX_TARGET_PLATFORM_NATIVE ON)
		endif()
	elseif("${NX_TARGET_PLATFORM}" IN_LIST NX_PLATFORMS_HAIKU)
		set(NX_TARGET_PLATFORM_STRING "Haiku")
		set(NX_TARGET_PLATFORM_HAIKU ON)
		set(NX_TARGET_PLATFORM_POSIX ON)
		if(NX_HOST_PLATFORM_HAIKU)
			set(NX_TARGET_PLATFORM_NATIVE ON)
		endif()
	elseif("${NX_TARGET_PLATFORM}" IN_LIST NX_PLATFORMS_LINUX)
		set(NX_TARGET_PLATFORM_STRING "Linux")
		set(NX_TARGET_PLATFORM_LINUX ON)
		set(NX_TARGET_PLATFORM_POSIX ON)
		if(NX_HOST_PLATFORM_LINUX)
			set(NX_TARGET_PLATFORM_NATIVE ON)
		endif()
	elseif("${NX_TARGET_PLATFORM}" IN_LIST NX_PLATFORMS_MSDOS)
		set(NX_TARGET_PLATFORM_STRING "MSDOS")
		set(NX_TARGET_PLATFORM_MSDOS ON)
	elseif("${NX_TARGET_PLATFORM}" IN_LIST NX_PLATFORMS_SOLARIS)
		set(NX_TARGET_PLATFORM_STRING "Solaris")
		set(NX_TARGET_PLATFORM_SOLARIS ON)
		set(NX_TARGET_PLATFORM_POSIX ON)
		if(NX_HOST_PLATFORM_SOLARIS)
			set(NX_TARGET_PLATFORM_NATIVE ON)
		endif()
	elseif("${NX_TARGET_PLATFORM}" IN_LIST NX_PLATFORMS_WINDOWS)
		set(NX_TARGET_PLATFORM_STRING "Windows")
		set(NX_TARGET_PLATFORM_WINDOWS ON)
		if(NX_HOST_PLATFORM_WINDOWS)
			set(NX_TARGET_PLATFORM_NATIVE ON)
		endif()
	else()
		message(FATAL_ERROR "Unsupported Platform Type ('${NX_TARGET_PLATFORM}')")
	endif()

	string(TOUPPER "KYUII_PLATFORM_${NX_TARGET_PLATFORM_STRING}" NX_TARGET_PLATFORM_MACRO)
	list(APPEND NX_TARGET_PLATFORM_MACRO KYUII_PLATFORM_STRING="${NX_TARGET_PLATFORM_STRING}")
	
	if(NX_TARGET_PLATFORM_POSIX)
		list(APPEND NX_TARGET_PLATFORM_MACRO "NX_TARGET_PLATFORM_POSIX")
	endif()

	if(NOT NX_TARGET_PLATFORM_NATIVE)
		set(NX_TARGET_PLATFORM_CROSS ON)
	endif()

	nx_scope_propagate()
endfunction()

function(nx_identify_target)
	nx_scope_initialize()
	nx_identify_build()
	nx_identify_architecture()
	nx_identify_platform()
	nx_scope_propagate()
endfunction()
